/*----------------------------------------------------

Virtual Acoustic display demo.  The speech is read from
a monaural speech file and split into two channels.
Front-back discrimination is a factor derived from a
synthetic filter which gives +12 dB of gain at 0 deg.
azimuth and -12 dB at 180 deg., and follows an 
exp(cosine(azimuth)) distribution.  Side to side 
discrimination is a factor applied to the shadowed ear 
only using the difference in the frequency response 
between the shadowed and unshadowed ear derived from 
HRTFs measured for the Kemar head model.  First 
order room reflections from a room model are used to 
externalize the sound.

The room dimensions (drag window), source position (left 
mouse button) and head position (right mouse button) are 
user defineable.  The left mouse button also gives 
information on the size of the left and right ear filters 
applied to the sound.
 

Karen Bryden 
Manager, Voice Systems and Technologies 


E-mail: karen.bryden@crc.ca



         (c) Copyright Karen Bryden, 2000


This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

See File COPYING.txt


----------------------------------------------------*/

//includes

#include "3Dsound.h"
#include "glue.h"

//local prototypes
LRESULT CALLBACK WndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
int getData2 (FILE *sourceFile, short *inputData, int nSamples);

//external variables
extern POINT topPt, bottomPt, leftPt, rightPt;
extern struct compositeFilter filterL, filterR;
extern float fAzimuth, fRange;

//global variables
HWAVEOUT hWaveOut = NULL;
WAVEFORMATEX inputFormat, outputFormat;//32 bit

LPWAVEHDR lpOutHdr[numOutBufs];
HGLOBAL hGlobalOutBuf[numOutBufs];
HGLOBAL hGlobalOutHdr[numOutBufs];

int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance,
          LPSTR lpszCmdLine, int nCmdShow)
{
   static char szAppName[] = "3Daudio" ;
   HWND    hwnd ;
   MSG     msg ;
   WNDCLASS  wndclass ;

   if (!hPrevInstance) 
     {
     wndclass.style     = CS_HREDRAW | CS_VREDRAW ;
     wndclass.lpfnWndProc  = WndProc ;
     wndclass.cbClsExtra  = 0 ;
     wndclass.cbWndExtra  = 0 ;
     wndclass.hInstance   = hInstance ;
     wndclass.hIcon     = LoadIcon (NULL, IDI_APPLICATION) ;
     wndclass.hCursor    = LoadCursor (NULL, IDC_ARROW) ;
     wndclass.hbrBackground = GetStockObject (WHITE_BRUSH) ;
     wndclass.lpszMenuName = "fileMenu";
     wndclass.lpszClassName = szAppName ;

     RegisterClass (&wndclass) ;
     }
   hwnd = CreateWindow (szAppName, "3D Soundscape",
             WS_OVERLAPPEDWINDOW,
             20, 20,//top left corner
             600, 440,//width x height
             NULL, NULL, hInstance, NULL) ;

   ShowWindow (hwnd, nCmdShow) ;
   UpdateWindow (hwnd) ;

   while (GetMessage (&msg, NULL, 0, 0))
     {
     TranslateMessage (&msg) ;
     DispatchMessage (&msg) ;
     }

return msg.wParam ;
}
//               ***********************************
//               *  s t a r t  o f  w n d P r o c  *
//               ***********************************

LRESULT CALLBACK WndProc (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
static RECT roomModel, head, nose;
//static POINT headCenter = {180, 160}, sourcePt = {180, 160};set by WM_SIZE
static POINT headCenter, sourcePt;
static short cxBlock, cyBlock, cyText;
static int  cxChar, cyChar, numInDevs, numOutDevs;
static OPENFILENAME fileStruct;
static FILE *pSourceFile = NULL;
static char szFilter[] = 
{"Raw Speech (*.PCM)\0*.pcm\0"};
static char szFile[256], szFileTitle[256];

HDC     hdc ;
PAINTSTRUCT ps ;
TEXTMETRIC  tm ;
int returnValue, i;
char szErrorString[120];
short sBuffer[inBufSize];

switch (message)
{

case WM_CREATE:
hdc = GetDC (hwnd) ;
SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
GetTextMetrics (hdc, &tm) ;
cxChar = tm.tmAveCharWidth ;
cyChar = tm.tmHeight + tm.tmExternalLeading ;
ReleaseDC (hwnd, hdc) ;

//initialize the soundscape tables and allocate two buffers of size inBufSize
soundscapeInit(inBufSize);


//                      ***************************
//                      *  a u d i o   s e t u p  *
//                      ***************************
//specify output sampling parameters
//Win98 compatible
    inputFormat.wFormatTag       = WAVE_FORMAT_PCM;
    inputFormat.nChannels        = NChannelsIn;
    inputFormat.nSamplesPerSec   = SRateIn;
    inputFormat.nAvgBytesPerSec  = SRateIn * NChannelsIn * BytesPerSampleIn;
    inputFormat.nBlockAlign      = NChannelsIn * BytesPerSampleIn;
    inputFormat.wBitsPerSample      = BytesPerSampleIn * 8;
    outputFormat.wFormatTag      = WAVE_FORMAT_PCM;
    outputFormat.nChannels       = NChannelsOut;
    outputFormat.nSamplesPerSec  = SRateOut;    
    outputFormat.nAvgBytesPerSec = SRateOut * NChannelsOut * BytesPerSampleOut;    //required average data transfer rate
    outputFormat.nBlockAlign     = NChannelsOut * BytesPerSampleOut;             //1 channel * 1 bytes per sample
    outputFormat.wBitsPerSample     = BytesPerSampleOut * 8;

//test the audio I/O capability
numInDevs = waveInGetNumDevs();
numOutDevs = waveOutGetNumDevs();


//create buffers for the Output data
for (i = 0; i < numOutBufs; i++)
 {
 hGlobalOutHdr[i] = NULL;
 if((hGlobalOutHdr[i] = //create handle to memory of size WAVEHDR
 GlobalAlloc(GMEM_MOVEABLE|GMEM_SHARE, (DWORD)sizeof(WAVEHDR))) == NULL)
  {
  sprintf(szErrorString, "Out of memory");
  MessageBox (hwnd, szErrorString, "for OutHdr", MB_ICONEXCLAMATION | MB_OK);
  }
 else//handle becomes pointer
  {
  if((lpOutHdr[i] = (LPWAVEHDR)GlobalLock(hGlobalOutHdr[i])) == NULL)
   {
   sprintf(szErrorString, "GlobalLock error");
   MessageBox (hwnd, szErrorString, "for lpOutHdr", MB_ICONEXCLAMATION | MB_OK);
   }
  hGlobalOutBuf[i] = NULL;
  if((hGlobalOutBuf[i] =
   GlobalAlloc(GMEM_MOVEABLE|GMEM_SHARE, (DWORD)outBufSize)) == NULL)
   {
   sprintf(szErrorString, "Out of memory");
   MessageBox (hwnd, szErrorString, "for OutBuf", MB_ICONEXCLAMATION | MB_OK);
   }
  else //fill the header with the information about the buffer
   {
   if((lpOutHdr[i]->lpData = (LPSTR)GlobalLock(hGlobalOutBuf[i])) == NULL)
    {
    sprintf(szErrorString, "buffer %d GlobalLock error", i);
    MessageBox (hwnd, szErrorString, "for lpOutData", MB_ICONEXCLAMATION | MB_OK);
    }
   else
    {
    lpOutHdr[i]->dwBufferLength = outBufSize;
    lpOutHdr[i]->dwFlags = 0L;//in Win3.1 causes GPF if left unspecified!
    lpOutHdr[i]->dwLoops = 0L;//play only once
    }
   }
  }
 }

//try to open the output device
if(numOutDevs)//output device exists
 {
 if (returnValue = waveOutOpen(
  &hWaveOut,                  //Far *handle identifying opened waveform output device
  0,             //waveform output device to open
  &outputFormat,   //desired format for playback
  (DWORD)hwnd,                     //handle of window to process messages32 bit
  0,                       //only used with callback functions 32 bit
  CALLBACK_WINDOW))          //we have passed a handle to a window
  {
sprintf(szErrorString, "can't open audio output device, error %d", returnValue);
MessageBox (hwnd, szErrorString, "playback format not supported", MB_ICONEXCLAMATION | MB_OK);
  }
 else
  {
//prepare the output headers
  for(i = 0; i < numOutBufs; i++)
   {
   if(returnValue = waveOutPrepareHeader(
   hWaveOut,                   //handle of waveform Output device
   lpOutHdr[i],           //pointer to header of buffer 1
   (UINT)sizeof(WAVEHDR)))
    {
    sprintf(szErrorString, "buffer %d Error code %d", i, returnValue);
    MessageBox (hwnd, szErrorString, "error preparing OutHdr", MB_ICONEXCLAMATION | MB_OK);
    }
   else lpOutHdr[i]->dwFlags = lpOutHdr[i]->dwFlags | WHDR_DONE;//mark the buffer as done
   }
  }
 }
return 0 ;//end of WM_CREATE

//              *********************************************
//              * RBUTTONUP defines position of head        *
//              * and places sound source in middle of head *
//              *********************************************
case WM_RBUTTONUP:
 if((LOWORD(lParam) < roomModel.left)||
    (LOWORD(lParam) > roomModel.right)||
    (HIWORD(lParam) < roomModel.top)||
    (HIWORD(lParam) > roomModel.bottom))
  {
 returnValue = sprintf(szErrorString, 
  "Please click inside the room!                                     ");
 hdc = GetDC (hwnd) ;
 SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
 TextOut (hdc, cxChar, cyChar * (cyBlock / cyChar - 1),
 szErrorString, returnValue);
 ReleaseDC (hwnd, hdc) ;
 return 0;
  }
 headCenter.x = LOWORD(lParam);
 headCenter.y = HIWORD(lParam);
 head.top = headCenter.y - 11;
 head.bottom = headCenter.y + 11;
 head.left = headCenter.x - 13;
 head.right = headCenter.x + 13;
 nose.top = headCenter.y  - 3;
 nose.bottom = headCenter.y + 3;
 nose.left = head.right - 5;
 nose.right = head.right + 4;

//update the soundscape
soundscapeUpdate(&sourcePt, &headCenter, &roomModel, inBufSize);

InvalidateRect(hwnd, &roomModel, FALSE);//causes new head position to be displayed

return 0;


//              ************************************************
//              * LBUTTONUP defines the position of the source *
//              * and updates the right and left ear filters   *
//              ************************************************
//headCenter.y is the y axis center of the head
//headCenter.x is the x axis center of the head
//HIWORD(lParam) and LOWORD(lParam) define the y and x positions
//of the cursor

case WM_LBUTTONUP:
//calculate azimuth and distance of source
 if((LOWORD(lParam) < roomModel.left)||
    (LOWORD(lParam) > roomModel.right)||
    (HIWORD(lParam) < roomModel.top)||
    (HIWORD(lParam) > roomModel.bottom))
  {
 returnValue = sprintf(szErrorString, 
  "Please click inside the room!                                     ");
 hdc = GetDC (hwnd) ;
 SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
 TextOut (hdc, cxChar, cyChar * (cyBlock / cyChar - 1),
 szErrorString, returnValue);
 ReleaseDC (hwnd, hdc) ;
 return 0;
  }
//calculate azimuth and range of source relative to listener
 sourcePt.x = LOWORD(lParam);
 sourcePt.y = HIWORD(lParam);

//update the soundscape
soundscapeUpdate(&sourcePt, &headCenter, &roomModel, inBufSize);

//report filter details to user
 returnValue = sprintf(szErrorString, 
"Azim. %d, range %3.1f, L(%d taps, last %d), R(%d taps, last %d)          ",
  (int) (fAzimuth * 180.0f / 3.14159265f), //in degrees
  fRange, //in pixels
  filterL.totalTaps, 
  filterL.tap[filterL.totalTaps - 1].delay,
  filterR.totalTaps, 
  filterR.tap[filterR.totalTaps - 1].delay);
 hdc = GetDC (hwnd) ;
 SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
 TextOut (hdc, cxChar, cyChar * (cyBlock / cyChar - 1),
 szErrorString, returnValue);
 ReleaseDC (hwnd, hdc) ;
 InvalidateRect(hwnd, &roomModel, FALSE);//causes sound rays to be displayed

return 0;

//              *********************************************
//              *   WM_COMMAND responds to file operations  *
//              *********************************************

case WM_COMMAND:
 switch (wParam)
 {
 case fileOpen:
//file structure initialization
 memset(&fileStruct, '\0', sizeof (OPENFILENAME));//initialize structure to zero
 fileStruct.lStructSize       = sizeof (OPENFILENAME) ;
 fileStruct.hwndOwner         = hwnd ;
 fileStruct.lpstrFilter       = szFilter;
 fileStruct.nFilterIndex      = 1 ;//use first filter pair
 fileStruct.lpstrFile         = szFile;
 fileStruct.nMaxFile          = sizeof(szFile);
 fileStruct.lpstrFileTitle    = szFileTitle;
 fileStruct.nMaxFileTitle     = sizeof(szFileTitle);
 fileStruct.Flags             = OFN_FILEMUSTEXIST ;
 szFile[0] = '\0';//don't initialize filename edit box

//ask user to select file
 if(GetOpenFileName(&fileStruct)) //returns TRUE if file selected
  {
//if a file is already open, close it
  if(pSourceFile != NULL)fclose(pSourceFile);
//open the selected file
  pSourceFile = fopen( szFile, "rb");
  if ( pSourceFile == NULL)
   {
   sprintf(szErrorString, "can't open source file");
   MessageBox (hwnd, szErrorString, szFileTitle, MB_ICONEXCLAMATION | MB_OK);
   return 0;
   }
  }
 return 0;

 case filePlay:
 if(pSourceFile != NULL)
  {
/*
Reset the output device to return the buffers to the application
prior to starting playback before another file is completed.
note: This was eliminated because it sets the WHDR_DONE flag of all 
the buffers, even  though they may still be playing, WaveOutWrite 
to fail with "something still playing".  There is no need to reset 
the output device provided the WHDR_DONE flag is tested prior to 
loading new data into the output buffers.  The getdata calls simply 
continue using the new file specification.  See Windows 3.1 SDK,
multimedia programmers guide, chapter 5, "Managing audio data blocks".

  if(returnValue = waveOutReset(hWaveOut))
   {
   sprintf(szErrorString, "Error code %d", returnValue);
   MessageBox (hwnd, szErrorString, "can't stop audio Output dev", MB_ICONEXCLAMATION | MB_OK);
   }
*/
//rewind the file
  if(fseek(pSourceFile,0L,SEEK_SET) != 0)
   {
   sprintf(szErrorString, "can't rewind source file");
   MessageBox (hwnd, szErrorString, szFileTitle, MB_ICONEXCLAMATION | MB_OK);
   return 0;
   }
//load the output buffers
  for(i = 0; i < numOutBufs; i++)
   {
//check that the buffer is available before writing data
   if(lpOutHdr[i]->dwFlags & WHDR_DONE)
    {
//read data from the file
    if(getData2 (pSourceFile, sBuffer, inBufSize))
     {

//convert the monaural array of samples to 3D sound
spatialize(sBuffer, (short *)(lpOutHdr[i]->lpData), inBufSize);

//send buffer to the output device
     if(returnValue = waveOutWrite(
       hWaveOut,              //handle of waveform Output device
       lpOutHdr[i],           //pointer to header of buffer
       (UINT)sizeof(WAVEHDR)))
      {
      i = sprintf(szErrorString, 
 "waveOutWrite error, returned %d                 ", returnValue);
      hdc = GetDC (hwnd) ;
      SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
      TextOut (hdc, cxChar, cyChar * (cyBlock / cyChar - 1),
      szErrorString, i);
      ReleaseDC (hwnd, hdc) ;
      }
     }//end of file test
    }//buffer available test
   }//all available buffers loaded
  }//end of sourceFile open test
 return 0;
 }//end of switch on wParam
return 0;//end of case WM_COMMAND:

case MM_WOM_DONE:
//when buffers are returned, refill them until eof reached
//lParam is a far pointer to the header identifying buffer being returned

//read data from the file
   if(getData2 (pSourceFile, sBuffer, inBufSize))
    {

//convert the monaural array of samples to 3D sound
spatialize(sBuffer, (short *)(((LPWAVEHDR)lParam)->lpData), inBufSize);

//return buffer to the output device
   if(returnValue = waveOutWrite(
     hWaveOut,              //handle of waveform Output device
     (LPWAVEHDR)lParam,           //pointer to header of buffer
     (UINT)sizeof(WAVEHDR)))
    {
    i = sprintf(szErrorString,
 "waveOutWrite error, returned %d                 ", returnValue);
    hdc = GetDC (hwnd) ;
    SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
    TextOut (hdc, cxChar, cyChar * (cyBlock / cyChar - 1),
    szErrorString, i);
    ReleaseDC (hwnd, hdc) ;
    }
   }
return 0;

case WM_DESTROY:

//free memory allocated by soundscapeInit()
soundscapeFree();

//if a file is open, close it
if(pSourceFile != NULL)returnValue = fclose (pSourceFile);
//  if (returnValue != 0) doesn't work, don't know why
//   {
//   sprintf(szErrorString, "can't close source file");
//   MessageBox (hwnd, szErrorString, szFileTitle, MB_ICONEXCLAMATION | MB_OK);
//   return 0;
//   }

//if the audio output device is running, reset and close it
if(hWaveOut != NULL)
 {
//must return all pending buffers to app before closing Output device
 if(returnValue = waveOutReset(hWaveOut))
  {
  sprintf(szErrorString, "Error code %d", returnValue);
  MessageBox (hwnd, szErrorString, "can't stop audio Output dev", MB_ICONEXCLAMATION | MB_OK);
  }

//unprepare the headers of the buffers (while hWaveOut is valid)
 for(i = 0; i < numOutBufs; i++)
  {
  if(returnValue = waveOutUnprepareHeader(
  hWaveOut,                   //handle of waveform Output device
  lpOutHdr[i],           //pointer to header of buffer
  (UINT)sizeof(WAVEHDR)))
   {
   sprintf(szErrorString, "buffer %d Error code %d", i, returnValue);
   MessageBox (hwnd, szErrorString, "error unpreparing OutHdr", MB_ICONEXCLAMATION | MB_OK);
   }
  }

//close the output device
 if(returnValue = waveOutClose(hWaveOut))
  {
  sprintf(szErrorString, "Error code %d", returnValue);
  MessageBox (hwnd, szErrorString, "can't close audio output dev", MB_ICONEXCLAMATION | MB_OK);
  }
 }

//release the memory buffers
for(i = 0; i < numOutBufs; i++)
 {
 if(hGlobalOutBuf[i] != NULL)
  {
  if(returnValue = GlobalUnlock(hGlobalOutBuf[i]))
   {
   sprintf(szErrorString, "memory leak %d", returnValue);
   MessageBox (hwnd, szErrorString, "can't unlock hGlobalOutBuf", MB_ICONEXCLAMATION | MB_OK);
   }
  else
   {
   if(GlobalFree(hGlobalOutBuf[i]) != NULL)
    {
    sprintf(szErrorString, "memory leak");
    MessageBox (hwnd, szErrorString, "can't free hGlobalOutBuf", MB_ICONEXCLAMATION | MB_OK);
    }
   }
  }
 }

for(i = 0; i < numOutBufs; i++)
 {
  if(hGlobalOutHdr[i] != NULL)
  {
  if(returnValue = GlobalUnlock(hGlobalOutHdr[i]))
   {
   sprintf(szErrorString, "memory leak %d", returnValue);
   MessageBox (hwnd, szErrorString, "can't unlock hGlobalOutHdr", MB_ICONEXCLAMATION | MB_OK);
   }
  else
   {
   if(GlobalFree(hGlobalOutHdr[i]) != NULL)
    {
    sprintf(szErrorString, "memory leak");
    MessageBox (hwnd, szErrorString, "can't free hGlobalOutHdr", MB_ICONEXCLAMATION | MB_OK);
    }
   }
  }
 }

// close the parent window
 PostQuitMessage (0) ;
 return 0 ;

case WM_SIZE:
 cxBlock = LOWORD (lParam);//width of client area
 cyBlock = HIWORD (lParam);//height of client area
 cyText = cyBlock - 2 * cyChar;//height of text I/O area
 roomModel.left  = inset;
 roomModel.top  = inset;
 roomModel.right = cxBlock - inset;
 roomModel.bottom = cyText;
headCenter.x = (roomModel.left + roomModel.right) / 2;//place head in center of client area
headCenter.y = (roomModel.top + roomModel.bottom) / 2;
 head.top = headCenter.y - 11;
 head.bottom = headCenter.y + 11;
 head.left = headCenter.x - 13;
 head.right = headCenter.x + 13;
 nose.top = headCenter.y  - 3;
 nose.bottom = headCenter.y + 3;
 nose.left = head.right - 6;
 nose.right = head.right + 4;
 sourcePt = headCenter;
 topPt = headCenter;
 bottomPt = headCenter;
 leftPt = headCenter;
 rightPt = headCenter;
//soundscapeUpdate(&sourcePt, &headCenter, &roomModel, inBufSize);causes lines outside room
return 0 ;

case WM_PAINT:
hdc = BeginPaint (hwnd, &ps) ;

//SelectObject (hdc, GetStockObject (SYSTEM_FIXED_FONT)) ;
SetBkMode (hdc, TRANSPARENT) ;
Rectangle (hdc, roomModel.left, roomModel.top, roomModel.right, roomModel.bottom) ;//walls of room
Ellipse(hdc, head.left, head.top, head.right, head.bottom);//head
Arc(hdc, nose.left, nose.top, nose.right, nose.bottom,
    (nose.left + nose.right) / 2, nose.bottom,
    (nose.left + nose.right) / 2 - 1, nose.top);//nose
//show indirect rays
SelectObject (hdc, CreatePen(PS_DOT, 1, RGB(0,255,0)));
MoveToEx(hdc, topPt.x, topPt.y, NULL);
LineTo(hdc, headCenter.x, headCenter.y);
LineTo(hdc, leftPt.x, leftPt.y);
MoveToEx(hdc, rightPt.x, rightPt.y, NULL);
LineTo(hdc, headCenter.x, headCenter.y);
LineTo(hdc, bottomPt.x, bottomPt.y);
//show direct ray
DeleteObject(SelectObject (hdc, CreatePen(PS_SOLID, 1, RGB(0,0,255))));
MoveToEx(hdc, sourcePt.x, sourcePt.y, NULL);
LineTo(hdc, headCenter.x, headCenter.y);
//show source position
DeleteObject(SelectObject (hdc, CreatePen(PS_SOLID, 1, RGB(255,0,0))));
MoveToEx(hdc, sourcePt.x - 2, sourcePt.y, NULL);
LineTo(hdc, sourcePt.x + 3, sourcePt.y);
MoveToEx(hdc, sourcePt.x, sourcePt.y - 2, NULL);
LineTo(hdc, sourcePt.x, sourcePt.y + 3);
//restore default setting
DeleteObject(SelectObject (hdc, GetStockObject(BLACK_PEN)));

EndPaint (hwnd, &ps) ;//end of WM_PAINT
return 0 ;

}//end of switch on message

return DefWindowProc (hwnd, message, wParam, lParam) ;
}


int getData2 (FILE *sourceFile, short *inputData, int nSamples)
{
int result;
result = (int) fread(inputData, BytesPerSampleIn, nSamples, sourceFile);
if ( result == nSamples )return 1;
else return 0;
}

